/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:49:03 1991 by burrows        */
/* routines to handle an avl tree */

/* routines are:
 *	avl_add, avl_lookup, avl_iterate
 */
#include <stdio.h>
#include "avl.h"

extern char *emalloc ();
extern void free ();

#define SAMELENGTH	0
#define LEFTLONG	1
#define RIGHTLONG	2

static int addavl ();

/*
 * Add item to an avl tree.
 *
 * "ptree" is pointer to root of tree.
 * "cmp" is comparison function between data items
 * "data" is the data item to be added

 *
 * find position where "data" belongs in tree, and create a node for it if
 * it's not there
 *
 * returns pointer to position
 */
AVL_DATA *avl_add (ptree, cmp, data)
struct avl **ptree;
int (*cmp) ();
AVL_DATA *data;
{
	AVL_DATA *result;
	
	(void) addavl (ptree, cmp, data, &result);

	return (result);
}

static int addavl (ptree, cmp, data, result)
struct avl **ptree;
register int (*cmp) ();
AVL_DATA *data;
AVL_DATA **result;
{
	register struct avl *tree = *ptree, *save, *dsave;
	register int x;
	register int treegrew;

	if (tree) {
		x = AVL_CMP (cmp, data, &tree->data);

		if (x < 0) {
			if ((treegrew = addavl (&tree->left,cmp,data,result))>0) {
				switch (tree->balance)
				{
				case SAMELENGTH:
					tree->balance = LEFTLONG; 
					break;

				case LEFTLONG:
					save = tree->left;
					if (save->balance == LEFTLONG) {
						tree->left = save->right;
						save->right = tree;
						tree->balance = SAMELENGTH;
						*ptree = save;
					} else {
						dsave = save->right;
						save->right = dsave->left;
						dsave->left = save;
						tree->left = dsave->right;
						dsave->right = tree;
						switch (dsave->balance)
						{
						case SAMELENGTH:
							tree->balance = SAMELENGTH;
							save->balance = SAMELENGTH;
							break;

						case LEFTLONG:
							tree->balance = RIGHTLONG;
							save->balance = SAMELENGTH;
							break;
						
						case RIGHTLONG:
							tree->balance = SAMELENGTH;
							save->balance = LEFTLONG;
							break;
						}
						*ptree = dsave;
					}
					(*ptree)->balance = SAMELENGTH;
					treegrew = 0;
					break;

				case RIGHTLONG:
					tree->balance = SAMELENGTH;
					treegrew = 0;
					break;
				}
			}
		} else if (x > 0) {
			if ((treegrew = addavl (&tree->right,cmp,data,result))>0) {
				switch (tree->balance)
				{
				case SAMELENGTH:
					tree->balance = RIGHTLONG; 
					break;

				case RIGHTLONG:
					save = tree->right;
					if (save->balance == RIGHTLONG) {
						tree->right = save->left;
						save->left = tree;
						tree->balance = SAMELENGTH;
						*ptree = save;
					} else {
						dsave = save->left;
						save->left = dsave->right;
						dsave->right = save;
						tree->right = dsave->left;
						dsave->left = tree;
						switch (dsave->balance)
						{
						case SAMELENGTH:
							tree->balance = SAMELENGTH;
							save->balance = SAMELENGTH;
							break;

						case RIGHTLONG:
							tree->balance = LEFTLONG;
							save->balance = SAMELENGTH;
							break;
						
						case LEFTLONG:
							tree->balance = SAMELENGTH;
							save->balance = RIGHTLONG;
							break;
						}
						*ptree = dsave;
					}
					(*ptree)->balance = SAMELENGTH;
					treegrew = 0;
					break;

				case LEFTLONG:
					tree->balance = SAMELENGTH;
					treegrew = 0;
					break;
				}
			}
		} else {
			*result = &tree->data;
			treegrew = 0;		/* data already there */
		}
	} else {		/* we have to add at this level */
		tree = (struct avl *) emalloc (sizeof (struct avl));

		tree->left = tree->right = 0;
		tree->balance = SAMELENGTH;
		AVL_COPY (data, &tree->data);
		*ptree = tree;
		treegrew = 1;
		*result = &tree->data;
	}

	return (treegrew);
}

/* lookup item in an avl tree
 *
 * "ptree" is pointer to root of tree.
 * "cmp" is comparison function between data items
 * "data" is the data item to be deleted
 *
 * Find "data" in tree, and return data item in tree.
 * If it cannot be found, return null.
 *
 * returns
 *	&data
 *	null if data not in tree
 */
AVL_DATA *avl_lookup (ptree, cmp, data)
register struct avl **ptree;
register int (*cmp) ();
AVL_DATA *data;
{
	register struct avl *tree = *ptree;
	register int x;

	while (tree) {
		x = AVL_CMP (cmp, data, &tree->data);

		if (x < 0)
			tree = tree->left;
		else if (x > 0)
			tree = tree->right;
		else
			return (&tree->data);
	}
	return (0);
}

/* iterate over an avl tree
 *
 * "ptree" is pointer to root of tree.
 * "func" is a function to be called for each node in the tree
 *
 *   iterate over all nodes in in-order, calling   
 *
 *                   (*func) (data, node->data)
 */
int avl_iterate (ptree, func, data)
struct avl **ptree;
register int (*func) ();
register char *data;
{
	register struct avl *tree = *ptree;

	while (tree) {
		if (avl_iterate (&tree->left, func, data))
			return (1);

		if ((*func) (data, &tree->data))
			return (1);

		tree = tree->right;
	}
	return (0);
}

/* delete entire tree */
void avl_destroy (ptree)
struct avl **ptree;
{
	register struct avl *tree = *ptree;

	if (tree != 0) {
		if (tree->left != 0)
			avl_destroy (&tree->left);
		if (tree->right != 0)
			avl_destroy (&tree->right);

		AVL_DESTROY (&tree->data);
		free ((char *) tree);
		*ptree = 0;
	}
}
